/* 
 * Copyright (c) 2012 Fran Jagić
 * All rights reserved.
 * 
 * This file is part of JDosage2.
 * 
 * JDosage2 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * JDosage2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with JDosage2.  If not, see <http://www.gnu.org/licenses/>.
 */
package hr.fjagic.jdosage2;

import hr.fjagic.jdosage2.webcomic.Webcomic;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

/**
 * Command line interface for JDosage2.
 * 
 * Future versions will include parallel downloads, date/counter generated
 * webcomics and GUI.
 * 
 * @author Fran Jagi&#263;
 * @since 0.5
 */
public class JDosage2CLI {

	// cli options
	private static final String CLI_LIST = "list";
	private static final String CLI_DOWNLOAD = "download";
	// TODO set as system property
	private static final String CLI_IGNORE_MISSING = "ignoremissing";
	private static final String CLI_IGNORE_EXISTING = "ignoreexisting";

	/**
	 * Command line interface for JDosage2.
	 * 
	 * @param args
	 *            command line parameters
	 * @throws ClassNotFoundException
	 *             if selected logger is unavailable
	 */
	@SuppressWarnings("static-access")
	public static void main(String[] args) throws ClassNotFoundException {
		// logger initialization
		freemarker.log.Logger.selectLoggerLibrary(freemarker.log.Logger.LIBRARY_SLF4J);

		// cli option parsing
		Option download = OptionBuilder.withLongOpt(CLI_LIST).hasOptionalArgs().withArgName("webcomic...")
				.withDescription("Lists webcomics with definition files matching wildcard expression").create('l');
		Option list = OptionBuilder.withLongOpt(CLI_DOWNLOAD).hasArgs().withArgName("webcomic...")
				.withDescription("Downloads webcomics with definition files matching wildcard expression").create('d');
		Option ignoreExisting = OptionBuilder.withLongOpt(CLI_IGNORE_EXISTING)
				.withDescription("Don't stop the download when existing comic is found").create('e');
		Option ignoreMissing = OptionBuilder.withLongOpt(CLI_IGNORE_MISSING)
				.withDescription("Don't stop the download when no comics are found on a page").create('m');

		Options options = new Options();
		OptionGroup group = new OptionGroup();
		group.addOption(list);
		group.addOption(download);
		group.setRequired(true);
		options.addOptionGroup(group);
		options.addOption(ignoreExisting);
		options.addOption(ignoreMissing);

		try {
			CommandLine commandLine = new BasicParser().parse(options, args);

			Configuration config = Configuration.getInstance();
			WebcomicManager wcm = new WebcomicManager(config);
			if (commandLine.hasOption(CLI_LIST)) {
				String[] webcomicExps = commandLine.getOptionValues(CLI_LIST);
				if (webcomicExps == null || webcomicExps.length == 0) {
					webcomicExps = new String[] { "*", "*/*" };
				}
				Collection<Webcomic> webcomics = wcm.getWebcomics(webcomicExps);

				// sort by name
				List<Webcomic> sortedWebcomics = new ArrayList<Webcomic>(webcomics);
				Collections.sort(sortedWebcomics, new Comparator<Webcomic>() {

					@Override
					public int compare(Webcomic o1, Webcomic o2) {
						return o1.getName().compareTo(o2.getName());
					}

				});

				for (Webcomic webcomic : sortedWebcomics) {
					System.out.println(webcomic.getName());
				}
			} else if (commandLine.hasOption(CLI_DOWNLOAD)) {
				if (commandLine.hasOption(CLI_IGNORE_EXISTING)) {
					config.setIgnoreExistingComics(true);
				}
				if (commandLine.hasOption(CLI_IGNORE_MISSING)) {
					config.setIgnoreMissingComics(true);
				}

				String[] webcomicExps = commandLine.getOptionValues(CLI_DOWNLOAD);
				Collection<Webcomic> webcomics = wcm.getWebcomics(webcomicExps);
				wcm.downloadComics(webcomics);
			}
		} catch (ParseException e) {
			// invalid command line arguments
			System.out.println(e.getMessage());
			new HelpFormatter().printHelp("JDosage2 webcomic downloader", options);
		}
	}

}
